@@ -22,6 +22,7 @@ from liquid.ast import BlockNode
 
 from liquid.context import Context
 from liquid.expression import LoopExpression
+from liquid.expression import NIL
 from liquid.lex import tokenize_loop_expression
 
 from liquid.parse import expect
@@ -172,7 +173,7 @@ class TablerowNode(Node):
         name = self.expression.name
         loop_iter, length = self.expression.evaluate(context)
 
-        if self.expression.cols:
+        if self.expression.cols and self.expression.cols != NIL:
             cols = self.expression.cols.evaluate(context)
             assert isinstance(cols, int)
         else:
@@ -213,7 +214,7 @@ class TablerowNode(Node):
         name = self.expression.name
         loop_iter, length = await self.expression.evaluate_async(context)
 
-        if self.expression.cols:
+        if self.expression.cols and self.expression.cols != NIL:
             cols = await self.expression.cols.evaluate_async(context)
             assert isinstance(cols, int)
         else:
