@@ -24,6 +24,10 @@ from liquid.stream import TokenStream
 from liquid.parse import get_parser
 from liquid.utils import LRUCache
 
+from liquid.parse import parse_filtered_expression_value
+from liquid.parse import parse_boolean_expression_value
+from liquid.parse import parse_loop_expression_value
+
 from liquid import ast
 from liquid import builtin
 from liquid import loaders
@@ -52,6 +56,18 @@ class Environment:
     :param statement_end_string: The sequence of characters indicating the end of an
         output statement. Defaults to ``}}``
     :type statement_end_string: str
+    :param template_comments: If ``True``, enable template comments. Where, by default,
+        anything between ``{#`` and ``#}`` is considered a comment. Defaults to
+        ``False``.
+    :type template_comments: bool
+    :param comment_start_string: The sequence of characters indicating the start of a
+        comment. Defaults to ``{#``. ``template_comments`` must be ``True`` for
+        ``comment_start_string`` to have any effect.
+    :type comment_start_string: str
+    :param comment_end_string: The sequence of characters indicating the end of a
+        comment. Defaults to ``#}``. ``template_comments`` must be ``True`` for
+        ``comment_end_string`` to have any effect.
+    :type comment_end_string: str
     :param tolerance: Indicates how tolerant to be of errors. Must be one of
         ``Mode.LAX``, ``Mode.WARN`` or ``Mode.STRICT``. Defaults to ``Mode.STRICT``.
     :type tolerance: Mode
@@ -82,12 +98,15 @@ class Environment:
         Defaults to 300. If ``cache_size`` is ``None`` or less than ``1``, it has the
         effect of setting ``auto_reload`` to ``False``.
     :type cache_size: int
+    :param expression_cache_size: The capacity of each of the common expression caches.
+        Defaults to ``0``, disabling expression caching.
+    :type expression_cache_size: int
     :param globals: An optional mapping that will be added to the context of any
         template loaded from this environment. Defaults to ``None``.
     :type globals: dict
     """
 
-    # pylint: disable=redefined-builtin too-many-arguments
+    # pylint: disable=redefined-builtin too-many-arguments too-many-locals
     def __init__(
         self,
         tag_start_string: str = r"{%",
@@ -103,15 +122,36 @@ class Environment:
         auto_reload: bool = True,
         cache_size: int = 300,
         globals: Optional[Mapping[str, object]] = None,
+        template_comments: bool = False,
+        comment_start_string: str = "{#",
+        comment_end_string: str = "#}",
+        expression_cache_size: int = 0,
     ):
         self.tag_start_string = tag_start_string
         self.tag_end_string = tag_end_string
         self.statement_start_string = statement_start_string
         self.statement_end_string = statement_end_string
+
+        # Automatic tag stripping is not yet implemented. Chaning this has no effect.
         self.strip_tags = strip_tags
+
+        # An instance of a template loader implementing ``liquid.loaders.BaseLoader``.
+        # ``get_template()`` will delegate to this loader.
         self.loader = loader or loaders.DictLoader({})
+
+        # A mapping of template variable names to python objects. These variables will
+        # be added to the global namespace of any template rendered from this
+        # environment using ``from_string`` or ``get_template``.
         self.globals: Mapping[str, object] = globals or {}
 
+        # Extended template comment syntax control.
+        self.template_comments = template_comments
+        self.comment_start_string = comment_start_string
+        self.comment_end_string = comment_end_string
+        if not self.template_comments:
+            self.comment_start_string = ""
+            self.comment_end_string = ""
+
         # The undefined type. When an identifier can not be resolved, the returned value
         # is ``Undefined`` or a subclass of ``Undefined``.
         self.undefined = undefined
@@ -139,6 +179,16 @@ class Environment:
             self.cache = {}
             self.auto_reload = False
 
+        # Common expression parsing functions that might be cached.
+        self.expression_cache_size = expression_cache_size
+        self.parse_filtered_expression_value = parse_filtered_expression_value
+        self.parse_boolean_expression_value = parse_boolean_expression_value
+        self.parse_loop_expression_value = parse_loop_expression_value
+        self.set_expression_cache_size(expression_cache_size)
+
+        # Instances of ``template_class`` are returned from ``from_string``,
+        # ``get_template`` and ``get_template_async``. It should be the
+        # ``BoundTemplate`` class or a subslass of it.
         self.template_class = BoundTemplate
 
         builtin.register(self)
@@ -150,6 +200,8 @@ class Environment:
                 self.statement_end_string,
                 self.tag_start_string,
                 self.tag_end_string,
+                self.comment_start_string,
+                self.comment_end_string,
                 self.mode,
                 self.strip_tags,
                 # Necessary when replacing the standard output statement implementation.
@@ -192,6 +244,8 @@ class Environment:
             self.tag_end_string,
             self.statement_start_string,
             self.statement_end_string,
+            self.comment_start_string,
+            self.comment_end_string,
         )
         parser = get_parser(self)
 
@@ -206,6 +260,7 @@ class Environment:
         name: str = "",
         path: Optional[Union[str, Path]] = None,
         globals: Optional[Mapping[str, object]] = None,
+        matter: Optional[Mapping[str, object]] = None,
     ) -> BoundTemplate:
         """Parse the given string as a liquid template.
 
@@ -220,6 +275,9 @@ class Environment:
         :param globals: An optional mapping of context variables made available every
             time the resulting template is rendered. Defaults to ``None``.
         :type globals: dict
+        :param matter: Optional mapping containing variables associated with the
+            template. Could be "front matter" or other meta data.
+        :type matter: Optional[Mapping[str, object]]
         :returns: A parsed template ready to be rendered.
         :rtype: liquid.template.BoundTemplate
         """
@@ -235,6 +293,7 @@ class Environment:
             path=path,
             parse_tree=parse_tree,
             globals=self.make_globals(globals),
+            matter=matter,
         )
 
     # pylint: disable=redefined-builtin
@@ -305,10 +364,9 @@ class Environment:
     ) -> Dict[str, object]:
         """Combine environment globals with template globals."""
         if globals:
-            _globals = {**self.globals, **globals}
-        else:
-            _globals = {**self.globals}
-        return _globals
+            # Template globals take priority over environment globals.
+            return {**self.globals, **globals}
+        return dict(self.globals)
 
     def error(
         self,
@@ -316,7 +374,7 @@ class Environment:
         msg: Optional[str] = None,
         linenum: Optional[int] = None,
     ) -> None:
-        """Raise ,warn or ignore the given exception according to the current mode."""
+        """Raise, warn or ignore the given exception according to the current mode."""
         if not isinstance(exc, Error):
             exc = exc(msg, linenum=linenum)
         elif not exc.linenum:
@@ -327,8 +385,32 @@ class Environment:
         if self.mode == Mode.WARN:
             warnings.warn(str(exc), category=lookup_warning(exc.__class__))
 
+    def set_expression_cache_size(self, maxsize: int = 0) -> None:
+        """Create or replace cached versions of the common expression parsers. If
+        `maxsize` is less than ``1``, no expression caching will happen.
 
-# pylint: disable=redefined-builtin
+        :param maxsize: The maximum size of each expression cache.
+        :type maxsize: int
+        """
+        if maxsize >= 1:
+            self.parse_filtered_expression_value = lru_cache(maxsize=maxsize)(
+                parse_filtered_expression_value
+            )
+            self.parse_boolean_expression_value = lru_cache(maxsize=maxsize)(
+                parse_boolean_expression_value
+            )
+            self.parse_loop_expression_value = lru_cache(maxsize=maxsize)(
+                parse_loop_expression_value
+            )
+        else:
+            self.parse_filtered_expression_value = parse_filtered_expression_value
+            self.parse_boolean_expression_value = parse_boolean_expression_value
+            self.parse_loop_expression_value = parse_loop_expression_value
+
+        self.expression_cache_size = maxsize
+
+
+# pylint: disable=redefined-builtin too-many-arguments too-many-locals
 @lru_cache(maxsize=10)
 def get_implicit_environment(
     tag_start_string: str,
@@ -344,6 +426,10 @@ def get_implicit_environment(
     auto_reload: bool,
     cache_size: int,
     globals: Optional[Mapping[str, object]],
+    template_comments: bool,
+    comment_start_string: str,
+    comment_end_string: str,
+    expression_cache_size: int,
 ) -> Environment:
     """Return an :class:`Environment` initialized with the given arguments."""
     return Environment(
@@ -360,14 +446,18 @@ def get_implicit_environment(
         auto_reload=auto_reload,
         cache_size=cache_size,
         globals=globals,
+        template_comments=template_comments,
+        comment_start_string=comment_start_string,
+        comment_end_string=comment_end_string,
+        expression_cache_size=expression_cache_size,
     )
 
 
-# `Template` is a factory function masquerading as a class. The desire to have
-# an intuitive API and to please the static type checker outweighs this abuse of
-# Python naming conventions. At least for now.
+# ``Template`` is a factory function masquerading as a class. The desire to have an
+# intuitive API and to please the static type checker outweighs this abuse of Python
+# naming conventions. At least for now.
 
-# pylint: disable=redefined-builtin too-many-arguments invalid-name
+# pylint: disable=redefined-builtin too-many-arguments invalid-name too-many-locals
 def Template(
     source: str,
     tag_start_string: str = r"{%",
@@ -382,6 +472,10 @@ def Template(
     auto_reload: bool = True,
     cache_size: int = 300,
     globals: Optional[Mapping[str, object]] = None,
+    template_comments: bool = False,
+    comment_start_string: str = "{#",
+    comment_end_string: str = "#}",
+    expression_cache_size: int = 0,
 ) -> BoundTemplate:
     """Returns a :class:`liquid.template.BoundTemplate`, automatically creating an
     :class:`Environment` to bind it to.
@@ -403,6 +497,18 @@ def Template(
     :param statement_end_string: The sequence of characters indicating the end of an
         output statement. Defaults to ``}}``
     :type statement_end_string: str
+    :param template_comments: If ``True``, enable template comments. Where, by default,
+        anything between ``{#`` and ``#}`` is considered a comment. Defaults to
+        ``False``.
+    :type template_comments: bool
+    :param comment_start_string: The sequence of characters indicating the start of a
+        comment. Defaults to ``{#``. ``template_comments`` must be ``True`` for
+        ``comment_start_string`` to have any effect.
+    :type comment_start_string: str
+    :param comment_end_string: The sequence of characters indicating the end of a
+        comment. Defaults to ``#}``. ``template_comments`` must be ``True`` for
+        ``comment_end_string`` to have any effect.
+    :type comment_end_string: str
     :param tolerance: Indicates how tolerant to be of errors. Must be one of
         `Mode.LAX`, `Mode.WARN` or `Mode.STRICT`. Defaults to ``Mode.STRICT``.
     :type tolerance: Mode
@@ -429,6 +535,9 @@ def Template(
         Defaults to 300. If ``cache_size`` is ``None`` or less than ``1``, it has the
         effect of setting ``auto_reload`` to ``False``.
     :type cache_size: int
+    :param expression_cache_size: The capacity of each of the common expression caches.
+        Defaults to ``0``, disabling expression caching.
+    :type expression_cache_size: int
     :param globals: An optional mapping that will be added to the context of any
         template loaded from this environment. Defaults to ``None``.
     :type globals: dict
@@ -449,6 +558,10 @@ def Template(
         auto_reload=auto_reload,
         cache_size=cache_size,
         globals=None,
+        template_comments=template_comments,
+        comment_start_string=comment_start_string,
+        comment_end_string=comment_end_string,
+        expression_cache_size=expression_cache_size,
     )
 
     return env.from_string(source, globals=globals)
