@@ -5,18 +5,18 @@ from liquid.golden.case import Case
 cases = [
     Case(
         description="assign a filtered literal",
-        template="{% assign foo = 'foo' | upcase %}{{ foo }}",
+        template=r"{% assign foo = 'foo' | upcase %}{{ foo }}",
         expect="FOO",
     ),
     Case(
         description="local variables shadow global variables",
-        template="{{ foo }}{% assign foo = 'foo' | upcase %}{{ foo }}",
+        template=r"{{ foo }}{% assign foo = 'foo' | upcase %}{{ foo }}",
         expect="barFOO",
         globals={"foo": "bar"},
     ),
     Case(
         description="assign a range literal",
-        template="{% assign foo = (1..3) %}{{ foo }}",
-        expect="1..3",
+        template=r"{% assign foo = (1..3) %}{{ foo | join: '#' }}",
+        expect="1#2#3",
     ),
 ]
