@@ -12,10 +12,15 @@ from typing import Optional
 from typing import NamedTuple
 from typing import TYPE_CHECKING
 
+from liquid import ast
+
 from liquid import expression
 from liquid.expression import IdentifierPathElement
 
-from liquid import ast
+from liquid.lex import tokenize_boolean_expression
+from liquid.lex import tokenize_loop_expression
+from liquid.lex import tokenize_filtered_expression
+
 from liquid.stream import TokenStream
 
 from liquid.exceptions import LiquidSyntaxError
@@ -86,12 +91,11 @@ class Parser:
     def parse(self, stream: TokenStream) -> ast.ParseTree:
         """Parse the given stream of tokens into a tree."""
         root = ast.ParseTree()
+        statements = root.statements
 
         while stream.current.type != TOKEN_EOF:
             try:
-                node = self.parse_statement(stream)
-                if node:
-                    root.statements.append(node)
+                statements.append(self.parse_statement(stream))
             except Error as err:
                 self.env.error(err, linenum=stream.current.linenum)
 
@@ -134,13 +138,18 @@ class Parser:
         """Parse multiple nodes from a stream of tokens. Stop parsing nodes when we
         find a token in `end` or we reach the end of the stream."""
         block = ast.BlockNode(stream.current)
+        statements = block.statements
 
         while stream.current.type != TOKEN_EOF:
             if stream.current.type == TOKEN_TAG and stream.current.value in end:
                 break
             stmt = self.parse_statement(stream)
-            if stmt:
-                block.statements.append(stmt)
+            statements.append(stmt)
+            # Detect output nodes in any of this block's children. This is used by
+            # some tags to automatically suppress whitespace when no other output is
+            # present.
+            if stmt.force_output or getattr(stmt, "forced_output", False):
+                block.forced_output = True
             stream.next_token()
 
         return block
@@ -220,6 +229,17 @@ IDENTIFIER_TOKENS = frozenset(
     (TOKEN_IDENTIFIER, TOKEN_INTEGER, TOKEN_DOT, TOKEN_LBRACKET, TOKEN_STRING)
 )
 
+LOOP_TOKENS = frozenset(
+    (
+        TOKEN_LIMIT,
+        TOKEN_OFFSET,
+        TOKEN_COLS,
+        TOKEN_REVERSED,
+    )
+)
+
+FILTER_END_TOKENS = frozenset((TOKEN_PIPE, TOKEN_EOF))
+
 
 def parse_boolean(stream: TokenStream) -> expression.Boolean:
     """Read a boolean literal (true or false) from the token stream."""
@@ -382,13 +402,12 @@ def parse_unchained_identifier(
 # NOTE: These names are missleading. The args and options in question can apply to any
 # iterable in a loop expression, not just ranges.
 
-RangeArg = Optional[
-    Union[
-        expression.Identifier,
-        expression.IntegerLiteral,
-        expression.FloatLiteral,
-        expression.Continue,
-    ]
+RangeArg = Union[
+    expression.Identifier,
+    expression.IntegerLiteral,
+    expression.FloatLiteral,
+    expression.Continue,
+    expression.Nil,
 ]
 
 
@@ -432,11 +451,16 @@ def parse_range_option(stream: TokenStream) -> RangeOption:
     else:
         expect(stream, TOKEN_REVERSED)
         stream.next_token()  # Eat REVERSED
-        opt = RangeOption(tok=tok, arg=None)
+        opt = RangeOption(tok=tok, arg=expression.NIL)
 
     return opt
 
 
+def parse_loop_expression_value(value: str) -> expression.LoopExpression:
+    """Parse the given string as a loop expression."""
+    return parse_loop_expression(TokenStream(tokenize_loop_expression(value)))
+
+
 def parse_loop_expression(stream: TokenStream) -> expression.LoopExpression:
     """Parse a liquid loop expression, as one might find in a `for` tag.
 
@@ -465,12 +489,7 @@ def parse_loop_expression(stream: TokenStream) -> expression.LoopExpression:
     expr = expression.LoopExpression(name=name, iterable=iterable)
 
     # Optional range modifiers can appear in any order.
-    while stream.current.type in (
-        TOKEN_LIMIT,
-        TOKEN_OFFSET,
-        TOKEN_COLS,
-        TOKEN_REVERSED,
-    ):
+    while stream.current.type in LOOP_TOKENS:
         opt = parse_range_option(stream)
 
         if opt.tok.type == TOKEN_REVERSED:
@@ -570,7 +589,7 @@ class ExpressionParser:
         if stream.current.type == TOKEN_COLON:
             stream.next_token()
 
-            while stream.current.type not in (TOKEN_PIPE, TOKEN_EOF):
+            while stream.current.type not in FILTER_END_TOKENS:
                 if stream.peek.type == TOKEN_COLON:
                     # A named parameter
                     param_name = stream.current.value
@@ -628,6 +647,9 @@ class ExpressionParser:
         precedence: Precedence = Precedence.LOWEST,
     ) -> expression.Expression:
         """Parse a literal or logical expression."""
+        end_expression = self.END_EXPRESSION
+        infix_funcs = self.infix_funcs
+
         prefix = self.prefix_funcs.get(stream.current.type)
         if prefix is None:
             operator = reverse_operators.get(stream.current.type, stream.current.value)
@@ -636,10 +658,10 @@ class ExpressionParser:
         left = prefix(stream)
 
         while (
-            stream.peek.type not in self.END_EXPRESSION
+            stream.peek.type not in end_expression
             and precedence < self.peek_precedence(stream)
         ):
-            infix = self.infix_funcs.get(stream.peek.type)
+            infix = infix_funcs.get(stream.peek.type)
             if infix is None:
                 return left
 
@@ -648,6 +670,14 @@ class ExpressionParser:
 
         return left
 
+    def parse_filtered_expression_value(
+        self, value: str
+    ) -> expression.FilteredExpression:
+        """Parse the given value as a filtered expression."""
+        return self.parse_filtered_expression(
+            TokenStream(tokenize_filtered_expression(value))
+        )
+
     def parse_filtered_expression(
         self, stream: TokenStream
     ) -> expression.FilteredExpression:
@@ -661,6 +691,12 @@ class ExpressionParser:
         filters = self.parse_filters(stream)
         return expression.FilteredExpression(expr, filters)
 
+    def parse_boolean_expression_value(self, value: str) -> expression.Expression:
+        """Parse the given string as a boolean expression."""
+        return self.parse_boolean_expression(
+            TokenStream(tokenize_boolean_expression(value))
+        )
+
     def parse_boolean_expression(self, stream: TokenStream) -> expression.Expression:
         """Parse a liquid expression that evaluates to a Boolean value.
 
@@ -707,7 +743,13 @@ parse_prefix_expression = STANDARD_EXPRESSION_PARSER.parse_prefix_expression
 parse_infix_expression = STANDARD_EXPRESSION_PARSER.parse_infix_expression
 parse_expression = STANDARD_EXPRESSION_PARSER.parse_expression
 parse_filtered_expression = STANDARD_EXPRESSION_PARSER.parse_filtered_expression
+parse_filtered_expression_value = (
+    STANDARD_EXPRESSION_PARSER.parse_filtered_expression_value
+)
 parse_boolean_expression = STANDARD_EXPRESSION_PARSER.parse_boolean_expression
+parse_boolean_expression_value = (
+    STANDARD_EXPRESSION_PARSER.parse_boolean_expression_value
+)
 parse_assignment_expression = STANDARD_EXPRESSION_PARSER.parse_assignment_expression
 
 
